## ==========================================================================
## Retrieve unique identifier of a flowFrame
## (normally generated by the cytometer and normally unique)
## "$FIL" optional field replace by filename if missing
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("identifier", signature="flowFrame",
          definition=function(object){
            oid <- object@description[["GUID"]]
            if(is.null(oid) || is.na(oid))
               oid <- as.vector(object@description[["$FIL"]])
            if(is.null(oid) || is.na(oid))
                as.vector(object@description[["FILENAME"]])
            else
              as.vector(oid)
        })


## ==========================================================================
## accessor and replace methods for slot exprs
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("exprs", signature="flowFrame",
          definition=function(object)
            object@exprs
          )

setReplaceMethod("exprs", signature=c("flowFrame", "matrix"),
                 definition=function(object, value) {
                   object@exprs <- value
                   return(object)
                 })


## ==========================================================================
## accessor method for slot parameters
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("parameters", signature="flowFrame",
          definition=function(object)
            object@parameters
          )


## ==========================================================================
## accessor methods for items in the description slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("keyword",signature("flowFrame","character"), function(object,keyword) {
	structure(object@description[keyword],names=keyword)
})
setMethod("keyword",signature("flowFrame","function"), function(object,keyword) {
	keyword(object,object@description)
})
setMethod("keyword",signature("flowFrame","list"), function(object,keyword) {
	sapply(keyword,function(k) {
		if(is.character(k))
			object@description[k]
		else if(is.function(k))
			k(object,object@description,k)
		else NA
	})
})
setMethod("keyword",signature("flowFrame","missing"), function(object) {
    object@description
})

## ==========================================================================
## accessor methods for slot description
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("description",signature("flowFrame"),function(object) object@description)


## ==========================================================================
## accessor and replace methods for colnames of exprs slot
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("colnames", signature="flowFrame",
          definition=function(x, do.NULL="missing", prefix="missing")
          colnames(exprs(x))
          )

setReplaceMethod("colnames", signature=c("flowFrame", "ANY"),
                 definition=function(x, value) {
                   colnames(x@exprs) <- value
                   return(x)
                 })


## ==========================================================================
## accessor method for feature names
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("featureNames", signature="flowFrame",
          definition=function(object)
          object@parameters$desc
          )


## ==========================================================================
## accessor method for names
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("names", signature="flowFrame",
          definition=function(x) {
            cn <- colnames(x)
            fn <- featureNames(x)
            if(length(fn) == length(cn)) {
              cn <- paste("<", cn, ">", sep="")
              for(i in seq(along=fn)) {
                if(!is.na(fn[i]) && fn[i]!="")
                  cn[i] <- paste(cn[i],fn[i])
              }
            }
            cn
          })


## ==========================================================================
## a simple plot method without strange plot parameter and friends. It does
## the most intuitive thing: take a flowFrame and plot the first two columns
## of the data. If you want to plot other columns, subset the frame before
## plotting. For the more complicated stuff there may well be more elaborate
## methods but I advice to separate functionality. Let R do the method
## dispatching instead of putting everything into one "do-it-all" method.
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("plot", signature(x="flowFrame", y="missing"),
          definition=function(x, ...){
            values=exprs(x)
            geneplotter:::smoothScatter(values[,1:2], ...)
          })

setMethod("plot",signature(x="flowFrame",y="character"),function(x,y,...) {
	l = length(y)
	if(l==1)
		hist(exprs(x)[,y],...)
	else if(l==2) {
		geneplotter:::smoothScatter(exprs(x)[,y],...)
	}
})



## ==========================================================================
## the $-operator (subsetting with '$')
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"$.flowFrame" <- function(x, val)
    x[,val]


## ==========================================================================
## subsetting methods
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("[", signature="flowFrame",
          definition=function(x, i, j, ..., drop=FALSE) {
            exprs(x) <-  switch(1+missing(i)+2*missing(j),
                                { exprs(x)[i, j, ..., drop=drop] },
                                { exprs(x)[ , j, ..., drop=drop] },
                                { exprs(x)[i,  , ..., drop=drop] },
                                { exprs(x)[ ,  , ..., drop=drop] } )
            if(!missing(j)) 
              x@parameters = x@parameters[j,]
            x
          })
setMethod("[", signature=signature("flowFrame","filterResult"),
          definition=function(x,i,j,...,drop=FALSE) {
            if(missing(j))
              x[x %in% i,,...,drop=drop]
            else
              x[x %in% i,j,...,drop=drop]
          })
setMethod("[", signature=signature("flowFrame","filter"),
          definition=function(x,i,j,...,drop=FALSE) {
			result = filter(x,i)
            if(missing(j))
              x[result,,...,drop=drop]
            else
              x[result,j,...,drop=drop]
          })


## ==========================================================================
## nrow method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("nrow", signature=signature("flowFrame"),
          definition=function(x)
            return(nrow(x@exprs))
          )


## ==========================================================================
## ncol method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("ncol", signature=signature("flowFrame"),
          definition=function(x)
            return(ncol(x@exprs))
          )


## ==========================================================================
## show method for flowFrame
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("show",signature=signature("flowFrame"),
          definition=function(object) {
            dm <- dim(exprs(object))
            msg <- paste("flowFrame object with ", dm[1], " cells and ", 
                         dm[2], " observables:\n", paste(names(object), 
                         collapse = " "), "\nslot 'description' has ",
                         length(description(object)), " elements\n", sep = "")
            cat(msg)
            return(msg)
          })


## ==========================================================================
## Subset methods for flowFrame: Why is this Subset with capital 's' ???
## Why do we need the 'select' parameter? Wouldn't this be equivalent:
## Subset(x[,c(1,3)], subset)
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
setMethod("Subset", signature("flowFrame","filter"),
          definition=function(x,subset,select,...){
            if(!missing(select))
              Subset(x,x %in% subset,select,...)
            else
              Subset(x,x %in% subset,...)
          })
setMethod("Subset",signature("flowFrame","logical"),
          definition=function(x,subset,select,...) {
            if(!missing(select))
              x[subset & !is.na(subset),select]
            else x[subset,]
          })


## ==========================================================================
## split methods for flowFrame
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("split", signature("flowFrame","filter"),
          definition=function(x,f,drop=FALSE,...)
          split(x,filter(x,f),drop,...)
          )
#We actually filter on filterResults and multipleFilterResults
setMethod("split", signature("flowFrame","logicalFilterResult"),
          definition=function(x,f,drop=FALSE,population=NULL,prefix=NULL,flowSet=FALSE,...) {
	        if(is.null(population)) population=f@filterId
			if(!is.null(prefix)) paste(prefix,population,sep="")
            out = structure(list(x[f@subSet,],x[!f@subSet,]),
                      names=c(paste(population,"+",sep=""),
                        paste(population,"-",sep="")))
			if(length(flowSet) > 0 && flowSet) flowSet(out) else out
          })
setMethod("split", signature("flowFrame","multipleFilterResult"),
          function(x,f,drop=FALSE,prefix=NULL,flowSet=FALSE,...) {
            nn = if(is.null(prefix)) names(f) else paste(prefix,names(f),sep="")
            out = structure(lapply(seq(along=f),function(i) x[f[[i]],]),names=nn)
			if(length(flowSet) > 0 && flowSet) flowSet(out) else out
})
setMethod("split",signature("flowFrame","manyFilterResult"),function(x,f,drop=FALSE,prefix=NULL,flowSet=FALSE,...) {
	#If drop is TRUE then only use results without children
	nn  = if(drop) rownames(f@dependency)[rowSums(f@dependency)==0] else	names(f)
	out = structure(lapply(nn,function(i) x[f[[i]],]),names=if(is.null(prefix)) nn else paste(prefix,nn,sep=""))
	if(length(flowSet) > 0 && flowSet) {
		print(data.frame(name=nn,as.data.frame(f)[nn,],row.names=nn))
		flowSet(out,phenoData=new("AnnotatedDataFrame",
			data=data.frame(name=I(nn),as.data.frame(f)[nn,],row.names=nn),
			varMetadata=data.frame(labelDescription=I(c("Name","Filter")),row.names=c("name","filter"))))
	} else out
})
setMethod("split",signature("flowFrame","filterSet"),function(x,f,drop=FALSE,...) 
	split(x,filter(x,f),drop,...))
setMethod("split",signature("flowFrame","ANY"),function(x,f,drop=FALSE,prefix=NULL,...) 
	stop("invalid type for flowFrame split"))

## ==========================================================================
## summary method for flowFrame
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("summary", signature("flowFrame"), 
    function(object, ...) 
        apply(exprs(object), 2, summary)
 )


## ==========================================================================
## wrappers for apply (over exprs slot)
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("each_row",signature("flowFrame"),function(x,FUN,...) {
	apply(exprs(x),1,FUN,...)
})
setMethod("each_col",signature("flowFrame"),function(x,FUN,...) {
	apply(exprs(x),2,FUN,...)
})


## ===========================================================================
## spillover method
## ---------------------------------------------------------------------------
setMethod("spillover","flowFrame",function(x) {
	## Flow frames have a SPILL keyword that is often
	## used to store the spillover matrix. Attempt to
	## extract it.
	present <- unlist(keyword(x, c("spillover", "SPILL")))
        if(is.null(present)) stop("No spillover matrix stored in that flowFrame")
        else present
})


## ===========================================================================
## compensate method
## ---------------------------------------------------------------------------
setMethod("compensate",signature("flowFrame","matrix"),function(x,spillover) {
	## Make sure we're normalized to [0,1] and then invert
	cols = colnames(spillover)
	e    = exprs(x)
	e[,cols] = e[,cols]%*%solve(spillover/max(spillover))
	exprs(x) = e
	x
})


## ==========================================================================
## transform method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("transform",
          signature=signature(`_data`="flowFrame"),
          definition=function(`_data`, ...) {
              e <- substitute(list(...))
              x = `_data`
              transformed <- as.matrix(transform(as.data.frame(exprs(x)),...))
              ##Add any new parameter values
              if(ncol(transformed) > ncol(x@exprs)) {
              	cnames = c(colnames(x),colnames(transformed)[-c(1:ncol(x@exprs))])
              }
              else {
              	cnames = colnames(x)
              }
              
              colnames(transformed) = cnames
              new("flowFrame",
                  exprs=transformed, 
                  parameters=parameters(x),#[,params$name],
                  description=description(x))
          })


## ==========================================================================
## filter method
## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
setMethod("filter", signature(x="flowFrame", filter="filter"),
          function(x, filter) {
            result <- as(x %in% filter, "filterResult")
			identifier(result) = identifier(filter)
            filterDetails(result, identifier(filter)) <- filter
            result@frameId <- identifier(x)
            result
          })

setMethod("filter",signature(x="flowFrame",filter="filterSet"),function(x,filter) {
	# A list of filter names sorted by dependency
	fl = sort(filter,dependencies=TRUE)
	e  = filterSet()
	m  = as(filter,"list")
	r  = lapply(fl,function(n) {
		e[[n]] = eval(m[[n]])
		filter(x,e[n])
	})
	# A place to evaluate filters
#	e  = new.env()
#	assign(".__frame",x,env=e)
	# Evaluate each 
#	r = lapply(as(filter,"list")[fl],function(f) {
#		if(is(f,"formula")) f = eval(as.call(c(as.symbol("as"),f,"filter")),e)
#		print(f)
#		eval(as.call(c(as.symbol("filter"),as.symbol(".__frame"),f)),e)		
#	})
	if(all(sapply(r,is,"logicalFilterResult"))) {
		#Combine into a many filter result
		manyFilterResult(r,frameId=identifier(x),attr(fl,'AdjM'))
	} else 
		r
})
